# Things you can run to help support the overall effort
# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams and/or submitting predictions.

import requests, random
from pprint import pprint
from microprediction import new_key
import multiprocessing as mp

def donate(difficulty=None, password=None, donor='anonymous'):
    donate1(difficulty=difficulty,password=password, donor=donor)

def donaten(difficulty=None, password=None, donor='anonymous'):
    # Not stable enough
    num_procs = mp.cpu_count()
    pool = mp.Pool(8*num_procs)
    result = [ pool.apply(func=donate1,args=(difficulty,password,donor)) for _ in range(8*num_procs) ]
    pool.close()


def donate1(difficulty=None, password=None, donor='anonymous'):
    if password is None:
        try:
            from microprediction.config_private import DONATION_PASSWORD, DONOR_NAME
        except:
            raise Exception("Ask Peter for the donation password ")
    write_key = new_key(difficulty=8)
    res = requests.post('https://api.microprediction.org/donations/' + write_key,data={'password': password, 'donor': donor})
    if res.status_code==200:
        if 'password' in res.json()['message']:
            return {"error":res.json()}
        else:
            if difficulty is None:
                difficulty = random.choice([13])
            while True:
                print("Mining and donating the MUIDs with password "+password+" and donor name "+donor+". Thanks. Difficulty set to "+str(difficulty),flush=True)
                write_key = new_key(difficulty=12)
                print(write_key,flush=True)
                res = requests.post('https://api.microprediction.org/donations/' + write_key, data={'password': password,'donor':donor})
                pprint(res.json())
    else:
        return res


if __name__=="__main__":
    donate(difficulty=8,password='5581cee8a281f4fd8cbe404143017046',donor='test')



# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams (see the 4th module in particular)

import requests

URL        = "https://data.cityofnewyork.us/resource/i4gi-tjb9.json"  # MTA city of NY realtime feed
VERRAZANO  = "416"  # "Verrazano-Narrows-Bridge"
BRONX      = "142"  # "BE S Griswold - Castle Hill Avenue"

def fetch_live_data(key,field):
    r = requests.get(URL)
    if r.status_code==200:
        data = r.json()
        selection = [ d for d in data if int(d["id"])==int(key) and d["status"]=="0"]
        selection.sort(key=lambda x:x["data_as_of"], reverse=True)
        if len(selection)>0:
            record = selection[0]
            return record[field]
        else:
            return None
    else:
        return None

def verrazano_speed():
    return fetch_live_data(key=VERRAZANO, field="speed")

def bronx_speed():
    return fetch_live_data(key=BRONX, field="speed")


from microprediction.univariate.distmachine import LossDist
from abc import ABC
from collections import OrderedDict
from hyperopt import fmin, tpe, hp, STATUS_OK, STATUS_FAIL
from copy import deepcopy
from getjson import getjson
import json

# Distribution machine that can be fit (using hyperopt by default)

DEFAULT_HYPER_PARAMS = {'max_evals':15}


class FitDist(LossDist, ABC):

    def __init__(self, state: dict, params: OrderedDict, hyper_params:dict = None ):
        super().__init__(state=state, params=params)
        self.hyper_params = deepcopy(DEFAULT_HYPER_PARAMS)
        self.hyper_params.update(hyper_params)
        self.points_to_evaluate = list() # List of pretty good values from previous searches

    def fit(self, lagged_values, lagged_times=None):
        params_before = deepcopy(self.params)  # Playing safe
        state_before = deepcopy(self.state)

        best_params = self.hyperfit(lagged_values=lagged_values, lagged_times=lagged_times, **self.hyper_params)
        # Is it really better?
        if self.params is None:
            self.params.update(best_params)
            self.points_to_evaluate.append(best_params)
            changed = True
        else:
            loss_before = self.loss(lagged_values=lagged_values, lagged_times=lagged_times, params=params_before, state=state_before)
            loss_after  = self.loss(lagged_values=lagged_values, lagged_times=lagged_times, params=best_params, state=state_before)
            changed = loss_after < loss_before
            if loss_after < loss_before:
                self.params.update(best_params)
                self.points_to_evaluate.append(best_params)
            else:
                self.params = deepcopy(params_before)
        self.state = state_before
        return changed

    # Implementation...

    def _hyperloss(self, prms, lagged_values, lagged_times=None):
        """ Same as loss but with hyperopt friendly signature

              prms:    list of values of parameters

        """
        params = OrderedDict(dict([(k, v) for k, v in zip(self.params.keys(), prms)]))
        try:
            evaluation = self.loss(lagged_values=lagged_values, lagged_times=lagged_times,
                                   params=params, state=self.state)
            status = STATUS_OK
        except Exception:
            evaluation = -99999999
            status = STATUS_FAIL

        return {'loss': evaluation, 'status': status}

    def hyperfit(self, lagged_values: [float], lower_bounds: dict, upper_bounds: dict, lagged_times=None, space=None,
                 algo=None, max_evals=100 ):

        if space is None:
            if self.params is None:
                pass
            space = [hp.uniform(v, lower_bounds[v], upper_bounds[v]) for v in self.params]

        if algo is None:
            algo = tpe.suggest

        def fn(prms):
            return self._hyperloss(lagged_values=lagged_values, lagged_times=lagged_times, prms=prms)

        # Test a call to fn (easier stack trace)
        example_prms = list(self.params.values())
        test_value = fn(example_prms)

        max_points_to_evaluate = int(max_evals)/2
        if len(self.points_to_evaluate)>max_points_to_evaluate:
            self.points_to_evaluate = self.points_to_evaluate[-max_points_to_evaluate:]

        best_params = fmin(fn=fn, space=space, algo=algo, max_evals=max_evals,
                              points_to_evaluate=self.points_to_evaluate )
        return best_params








import numpy as np
from microconventions import sign_changes, is_process


# --------------------------------------------------------------------------
#            Monte Carlo
# --------------------------------------------------------------------------

def inv_cdf_walk(inv_cdf, k: int, x0: float) -> float:
    """ Generate sample from sum of independent increments generated by inverse cdf function
         inv_cdf  func   taking (0,1) --> R
         k        int    Number of steps
         x0       float  Starting value
    """
    # See crawler_examples/decastyle_cat for usage example
    x = x0
    for _ in range(k):
        u = np.random.rand()
        dx = inv_cdf(u)
        x = x + dx
    return x



from microprediction.univariate.fitdist import FitDist
from microprediction.sequentialcrawler import SequentialStreamCrawler
from typing import Type
from collections import OrderedDict
from pprint import pprint
from microprediction.samplers import is_process
import math
import numpy as np
import random
import time
import logging
from copy import deepcopy
from microprediction.univariate.arrivals import approx_dt
from microprediction.univariate.cdfvalues import nudged
from getjson import getjson
from microprediction.samplers import project_on_lagged_lattice
from microconventions.stats_conventions import is_discrete

# Stream crawler that periodically fits parameters, or loads them
# Video explaining this code at https://www.microprediction.com/fitcrawler


class FitCrawler(SequentialStreamCrawler):

    def __init__(self, write_key, machine_type: Type[FitDist], machine_state: dict = None,
                 machine_params: OrderedDict = None,
                 lower_bounds: dict = None, upper_bounds: dict = None, space=None, algo=None, max_evals=10,
                 min_seconds=20, min_elapsed=5 * 60 * 60, decay=0.02, param_base_url=None, **kwargs):
        """
                state          -  Dictionary of state held for each stream. Supply None unless restarting.
                params         -  Initial params of HyperDist objects
                lower_bounds, upper_bounds - An easy alternative to providing hyperopt 'space'
                space          -  Hyperopt space object
                algo           -  Hyperopt algorithm choice (supply None to let it choose)
                max_evals      -  Total number of trials to keep (=hyperopt function evaluations per fitting)
                min_seconds    -  Minimum number of seconds that must open up before we attempt a fitting
                min_elapsed    -  Minimum time between attempts to fit a given time series

        """
        self.machine_hyper_params = {'lower_bounds': lower_bounds, 'upper_bounds': upper_bounds,
                                     'space': space, 'algo': algo, 'max_evals': max_evals}

        super().__init__(write_key=write_key, machine_type=machine_type, machine_params=machine_params,
                         machine_state=machine_state, **kwargs)
        self.fit_queue = list()  # List of streams that need fitting
        self.min_seconds = min_seconds  # Minimum time that must open up in order that we try to fit
        self.last_fit_time = dict()
        self.min_elapsed = min_elapsed
        self.decay = decay
        self.param_base_url = param_base_url   # Can be used to get offline parameters

    def include_delay(self, delay=None, name=None, **ignore):
        # It isn't recommended to use FitCrawler for long term predictions
        # Override if you are feeling bold
        return delay < 1000

    def initial_state(self, name, lagged_values, lagged_times, machine_params=None, machine_state=None, **ignore):
        #
        machine = self.machine_type(params=deepcopy(machine_params),
                                    state=deepcopy(machine_state),
                                    hyper_params=deepcopy(self.machine_hyper_params))

        chronological_values = list(reversed(lagged_values))
        chronological_times = list(reversed(lagged_times))
        as_process = is_process(chronological_values)
        values = list(np.diff([0.] + chronological_values)) if as_process else chronological_values
        dts = list(np.diff([chronological_times[0] - 1.0] + chronological_times))

        for value, dt in zip(values, dts):
            machine.update(value=value, dt=dt)
        return {'t': lagged_times[0], 'machine': machine, 'as_process': True, 'dt': approx_dt(lagged_times),
                'name': name}

    def fit(self, name, lagged_values, lagged_times=None, **ignored):
        """ Fit one stream, but only if the distribution machine supplied has a fit method """
        # Might use offline fit
        changed = False
        if name not in self.stream_state:
            self.stream_state.update(
                {name: self.initial_state(name=name, lagged_values=lagged_values, lagged_times=lagged_times)})
        else:
            machine = self.stream_state[name]['machine']

            stored_params = None
            if self.param_base_url is not None:
                url = self.param_base_url + '/' + name
                try:
                    stored_params = getjson(url)
                    if stored_params is None:
                        machine.params = stored_params
                except:
                    print('Failed to get params from '+url)

            if stored_params is None:
                try:
                    changed = machine.fit(lagged_values=lagged_values, lagged_times=lagged_times)
                    if changed:
                        print('Found better params for ' + name)
                        pprint(machine.params)
                        print(' ', flush=True)
                except AttributeError:
                    print('Keeping the same params for ' + name)
                    pprint(machine.params)
                    print(' ', flush=True)

        return changed

    def sample(self, lagged_values, lagged_times=None, name=None, delay=None, **ignored):
        """ Override default sampler so that we put the stream in the fit queue """

        if is_process(lagged_values):
            if name not in self.stream_state:
                self.stream_state.update(
                    {name: self.initial_state(name=name, lagged_values=lagged_values, lagged_times=lagged_times,
                                              machine_state=self.machine_state, machine_params=self.machine_params)})
            state = self.stream_state[name]
            state = self.update_state(state=state, lagged_values=lagged_values, lagged_times=lagged_times)
            samples = self.sample_using_state(state=state, lagged_values=lagged_values, lagged_times=lagged_times,
                                              name=name, delay=delay)
            self.stream_state[name] = state
            if name not in self.fit_queue:
                self.fit_queue.append(name)
            return samples
        else:
            # Flee cases where it isn't really a process
            for delay in self.DELAYS:
                horizon = self.horizon_name(name=name, delay=delay)
                self.withdraw(horizon=horizon)

    def downtime(self, seconds, **ignored):
        """ Try to fit then put on the end of the queue again """
        start_time = time.time()
        name = None
        if seconds > self.min_seconds:
            try:
                name = self.fit_queue.pop(0)
            except IndexError:
                pass
        if name is not None:
            if name not in self.last_fit_time or time.time() - self.last_fit_time[name] > self.min_elapsed:
                try:
                    lagged_values, lagged_times = self.get_lagged_values_and_times(name=name)
                except AttributeError:  # backward compat
                    lagged_values = self.get_lagged_values(name=name)
                    lagged_times = self.get_lagged_times(name=name)
                if lagged_values is None or lagged_times is None:
                    import logging
                    logging.warning("Could not fit " + name + " because we failed to get lagged_values ")
                else:
                    self.fit(name=name, lagged_values=lagged_values, lagged_times=lagged_times)
                    self.fit_queue.append(name)
                    self.last_fit_time[name] = time.time()

        seconds_used = time.time() - start_time
        seconds_remaining = seconds - seconds_used
        if seconds_remaining > 1:
            super().downtime(seconds=seconds_remaining)

    def sample_using_state(self, state, lagged_values, lagged_times, name, delay, **ignored):
        """ Creates samples using previous walks """
        # Provides a default way to sample using state. Override as you see fit.

        chronological_values = list(reversed(lagged_values))
        chronological_times = list(reversed(lagged_times))
        chronological_dt = [1.0] + list(np.diff(chronological_times))
        dts = reversed(chronological_dt)
        num_steps = int(math.ceil(delay / state['dt']))

        machine = deepcopy(state['machine'])

        # Walk back to get a good place to start
        for value, dt in zip(lagged_values, dts):
            machine.update(value=value)

        # Walk forward tracking anchor
        walk = list()
        for value, dt in zip(chronological_values, chronological_dt):
            machine.update(dt=dt)
            machine.update(value=value)
            anchor = machine.state['anchor']
            walk.append(anchor)

        # Sample randomly from walk and noise distribution, with some recently weighting for the former
        measurement_noise = [machine.inv_cdf(p) for p in self.percentiles()]
        steps_back = range(num_steps + 1, len(lagged_values) - 1)
        weights = [ math.exp(-self.decay * lag) for lag in steps_back ]
        back_choices = random.choices(population=steps_back, weights=weights, k=self.num_predictions)
        num = len(lagged_values)
        samples = [walk[num - step_back + num_steps] - walk[num - step_back] + noise for step_back, noise in
                   zip(back_choices, measurement_noise)]

        if is_discrete(lagged_values=lagged_values, num=0.75*len(lagged_values), ndigits=5):
            return project_on_lagged_lattice(values=samples, lagged_values=lagged_values)
        else:
            return sorted(nudged(samples))







from microprediction.conventions import MicroConventions, new_key
from microprediction.reader import MicroReader
from microprediction.writer import MicroWriter
from microprediction.crawler import MicroCrawler
from microprediction.simplecrawler import SimpleCrawler, RegularCrawler
from microprediction.onlinecrawler import OnlineStreamCrawler, OnlineHorizonCrawler
from microprediction.statefulcrawler import StreamCrawler
from microprediction.sequentialcrawler import DistMachine, SequentialStreamCrawler
from microprediction.supporter import donate
from microprediction.looping import PandasLoop
from microprediction.defaultcrawler import DefaultCrawler
from microprediction.univariate.digestdist import DigestDist
from microprediction.univariate.normaldist import NormalDist
from microprediction.univariate.distmachine import DistMachine
from microprediction.univariate.skewdist import SkewDist
from microprediction.polling import MicroPoll, ChangePoll
DistributionMachine = DistMachine  # Backward compat



# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams (see the 4th module in particular)

import logging
import requests
import urllib
import time
import pytz
import random
from datetime import datetime
from pprint import  pprint
from microprediction import MicroWriter
from microprediction.config_private import TRAFFIC_WRITE_KEY_13
from apscheduler.schedulers.blocking import BlockingScheduler


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def wait_between_attempts():
    """ Incremental backoff between connection attempts """
    wait_time = 19.3  # time is in seconds
    while True:
        yield wait_time
        wait_time = min(wait_time * 1.5, 30)
        wait_time *= (100 + random.randint(0, 50)) / 100

wait_time = wait_between_attempts()


def fetch_live_data():
    while True:
        try:
            file = urllib.request.urlopen("https://www.ndbc.noaa.gov/data/realtime2/WPOW1.cwind")
            if file.status==200:
                # NOTE: file is not indexable
                for i, line in enumerate(file):
                    # most recent data is at line number 3 in the file
                    if i == 2:
                        data = line.decode("utf-8").split()
                        direction = float(data[5]) / 360        # normalize between [0, 1)
                        speed = float(data[6]) / 10             # attempt to normalize
                        return speed, direction
        except:
            logger.error("Connection error: reconnecting...")
            time.sleep(next(wait_time))
    return speed, direction


NAME_SPEED = 'seattle_wind_speed.json'
NAME_DIR   = 'seattle_wind_direction.json'

initial_speed, initial_dir = fetch_live_data()
print("Initial speed is " + str(initial_speed) + " m/s and initial direction is " + str(initial_dir), flush=True)

try:
    mw = MicroWriter(write_key=TRAFFIC_WRITE_KEY_13)
except:
    raise Exception("You need to set the write key for this example to work")


def poll_and_send():
    utc_now = pytz.utc.localize(datetime.utcnow())
    pst_now = utc_now.astimezone(pytz.timezone("America/Los_Angeles"))
    speed, direction = fetch_live_data()
    res = mw.cset(names=[NAME_SPEED, NAME_DIR],values=[speed, direction])
    pprint({'PST time':pst_now.strftime("%H:%M"),'speed':speed,'direction':direction,"res":res})
    print('',flush=True)

def run():
    print('Starting scheduler',flush=True)
    scheduler = BlockingScheduler()
    scheduler.add_job(poll_and_send, 'interval', minutes=1)
    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass
    print('Stopping scheduler',flush=True)


if __name__=="__main__":
    run()
